use std::collections::HashSet;

use futures::TryStreamExt;
use log::{error, info, trace};
use mongodb::{
    bson::{doc, from_bson, to_document, Bson, Document},
    Client as MongoClient, Collection, Database,
};

use crate::environment::get_env_variable;

use super::models::{YorokobotCollection, COLLECTIONS_NAMES};

/// Database client
pub struct Client {
    mongo_client: Option<MongoClient>,
    database: Option<Database>,
}

impl Client {
    /// Create a new database client
    pub fn new() -> Client {
        Client {
            mongo_client: None,
            database: None,
        }
    }

    pub fn get_database(&self) -> &Database {
        match &self.database {
            Some(db) => db,
            None => {
                error!("Tried to access to the database before instantiating it");
                panic!();
            }
        }
    }

    /// Connect the client
    pub async fn connect(&mut self) {
        self.mongo_client = match MongoClient::with_uri_str(get_env_variable("MONGODB_URI")).await {
            Ok(c) => {
                info!("Successfully connected to the Mongo database");
                Some(c)
            }
            Err(e) => {
                error!("Failed to connect to the Mongo database: {e:#?}");
                panic!();
            }
        };

        self.database = match &self.mongo_client {
            Some(c) => Some(c.database(get_env_variable("MONGODB_DATABASE").as_str())),
            None => {
                error!("Got an unexpected None from self.database");
                panic!();
            }
        };

        self.check_init_error().await;
    }

    async fn check_init_error(&mut self) {
        info!("Launching initial database checks");

        let database = self.get_database();

        Self::check_collections_presence(database).await;
    }

    async fn check_collections_presence(db: &Database) {
        trace!("Starting the collections presence check for the database");

        let mut missing_collections: Vec<&str> = vec![];
        let collections: HashSet<String> = match db.list_collection_names(None).await {
            Ok(n) => n.into_iter().collect(),
            Err(e) => {
                error!("Failed to get the collections for the database: {e:#?}");
                panic!();
            }
        };

        for col in COLLECTIONS_NAMES {
            if !collections.contains(col) {
                missing_collections.push(col);
            }
        }

        if !missing_collections.is_empty() {
            error!(
                "Missing the following collections in the Database: {}",
                missing_collections.join(", ")
            );
            panic!();
        }
    }

    #[allow(dead_code)]
    fn get_collection<T: YorokobotCollection>(&self) -> Collection<Document> {
        self.get_database().collection(&T::get_collection_name())
    }

    fn get_typed_collection<T: YorokobotCollection>(&self) -> Collection<T> {
        self.get_database()
            .collection::<T>(&T::get_collection_name())
    }

    pub async fn get_by_id<T: YorokobotCollection>(&self, id: &str) -> Result<Option<T>, ()> {
        self.get_one::<T>(doc! {"_id": id}).await
    }

    pub async fn get_one<T: YorokobotCollection>(&self, filter: Document) -> Result<Option<T>, ()> {
        match self
            .get_typed_collection::<T>()
            .find_one(filter, None)
            .await
        {
            Ok(e) => Ok(e),
            Err(_) => Err(()),
        }
    }

    #[allow(dead_code)]
    pub async fn get_all<T: YorokobotCollection>(
        &self,
        filter: Option<Document>,
    ) -> Result<Vec<T>, ()> {
        let mut matching_docs: Vec<T> = vec![];

        let result = match filter {
            Some(f) => self.get_collection::<T>().find(f, None).await,
            None => self.get_collection::<T>().find(doc! {}, None).await,
        };

        let mut cursor = match result {
            Ok(c) => c,
            Err(_) => return Err(()),
        };

        while let Some(document) = match cursor.try_next().await {
            Ok(e) => e,
            Err(_) => return Err(()),
        } {
            match from_bson(Bson::Document(document)) {
                Ok(d) => matching_docs.push(d),
                Err(_) => return Err(()),
            }
        }

        Ok(matching_docs)
    }

    // TODO: Set true error handling
    pub async fn insert_one<T: YorokobotCollection>(&self, document: T) -> Result<(), ()> {
        match self
            .get_typed_collection::<T>()
            .insert_one(document, None)
            .await
        {
            Ok(_) => Ok(()),
            Err(_) => Err(()),
        }
    }

    // TODO: Set true error handling
    #[allow(dead_code)]
    pub async fn insert_many<T: YorokobotCollection>(&self, documents: Vec<T>) -> Result<(), ()> {
        match self
            .get_typed_collection::<T>()
            .insert_many(documents, None)
            .await
        {
            Ok(_) => Ok(()),
            Err(_) => Err(()),
        }
    }

    // TODO: Set true error handling
    #[allow(dead_code)]
    pub async fn delete_one<T: YorokobotCollection>(&self, document: Document) -> Result<u64, ()> {
        match self
            .get_typed_collection::<T>()
            .delete_one(document, None)
            .await
        {
            Ok(result) => Ok(result.deleted_count),
            Err(_) => Err(()),
        }
    }

    // TODO: Set true error handling
    #[allow(dead_code)]
    pub async fn delete_by_id<T: YorokobotCollection>(&self, id: &str) -> Result<(), ()> {
        match self
            .get_typed_collection::<T>()
            .delete_one(doc! {"_id": id}, None)
            .await
        {
            Ok(_) => Ok(()),
            Err(_) => Err(()),
        }
    }

    // TODO: Set true error handling
    #[allow(dead_code)]
    pub async fn delete_many<T: YorokobotCollection>(&self, document: Document) -> Result<(), ()> {
        match self
            .get_typed_collection::<T>()
            .delete_many(document, None)
            .await
        {
            Ok(_) => Ok(()),
            Err(_) => Err(()),
        }
    }

    //TODO: Set true error handling
    pub async fn update_one<T: YorokobotCollection>(
        &self,
        object: T,
        update: Document,
    ) -> Result<(), ()> {
        let serialized_doc = match to_document(&object) {
            Ok(d) => d,
            Err(_) => return Err(()),
        };

        match self
            .get_typed_collection::<T>()
            .update_one(serialized_doc, update, None)
            .await
        {
            Ok(_) => Ok(()),
            Err(_) => Err(()),
        }
    }

    //TODO: Set true error handling
    #[allow(dead_code)]
    pub async fn update_by_id<T: YorokobotCollection>(
        &self,
        document_id: &str,
        update: Document,
    ) -> Result<(), ()> {
        match self
            .get_typed_collection::<T>()
            .update_one(doc! {"_id": document_id}, update, None)
            .await
        {
            Ok(_) => Ok(()),
            Err(_) => Err(()),
        }
    }

    //TODO: Set true error handling
    #[allow(dead_code)]
    pub async fn update_many<T: YorokobotCollection>(
        &self,
        document: Document,
        update: Document,
    ) -> Result<(), ()> {
        match self
            .get_typed_collection::<T>()
            .update_many(document, update, None)
            .await
        {
            Ok(_) => Ok(()),
            Err(_) => Err(()),
        }
    }
}
