use log::{info, warn};
use std::sync::Arc;

use serenity::{
    async_trait,
    model::prelude::{
        command::{Command, CommandOptionType},
        interaction::application_command::{
            ApplicationCommandInteraction, CommandDataOption, CommandDataOptionValue,
        },
    },
    prelude::Context,
};

use crate::database::Client as DatabaseClient;

#[derive(Debug)]
pub enum CommandExecutionError {
    ArgumentExtractionError(String),
    ArgumentDeserializationError(String),
    DatabaseQueryError(String),
    ContextRetrievalError(String),
    DiscordAPICallError(String),
    SelectorError(String),
    UnknownCommand(String),
}

pub struct BotCommandOption {
    pub name: String,
    pub description: String,
    pub kind: CommandOptionType,
    pub required: bool,
}

#[async_trait]
pub trait BotCommand {
    fn new(interaction: ApplicationCommandInteraction) -> Self;
    fn name() -> String;
    fn description() -> String;
    fn options_list() -> Vec<BotCommandOption>;
    async fn run(
        &self,
        context: Context,
        database: Arc<DatabaseClient>,
    ) -> Result<(), CommandExecutionError>;

    fn extract_option_value(
        options: &[CommandDataOption],
        index: usize,
    ) -> Option<CommandDataOptionValue> {
        let serialized_opt = options.get(index)?;
        serialized_opt.resolved.clone()
    }

    async fn register(context: &Context) {
        info!("Starting the {} command registration", Self::name());
        match Command::create_global_application_command(context, |command| {
            let mut new_command = command.name(Self::name()).description(Self::description());

            for opt in Self::options_list() {
                new_command = new_command.create_option(|option| {
                    option
                        .name(opt.name)
                        .description(opt.description)
                        .kind(opt.kind)
                        .required(opt.required)
                });
            }
            new_command
        })
        .await
        {
            Ok(_) => info!("Successfully registered the {} command", Self::name()),
            Err(_) => warn!("Failed to register the {} command", Self::name()),
        };
    }
}
